@@ -1,7 +1,7 @@
 /* Diff files from a tar archive.
 
    Copyright (C) 1988, 1992, 1993, 1994, 1996, 1997, 1999, 2000, 2001,
-   2003, 2004 Free Software Foundation, Inc.
+   2003, 2004, 2005 Free Software Foundation, Inc.
 
    Written by John Gilmore, on 1987-04-30.
 
@@ -17,27 +17,16 @@
 
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
-   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 #include <system.h>
 
-#if HAVE_UTIME_H
-# include <utime.h>
-#else
-struct utimbuf
-  {
-    long actime;
-    long modtime;
-  };
-#endif
-
 #if HAVE_LINUX_FD_H
 # include <linux/fd.h>
 #endif
 
-#include <quotearg.h>
-
 #include "common.h"
+#include <quotearg.h>
 #include <rmt.h>
 #include <stdarg.h>
 
@@ -56,6 +45,8 @@ diff_init (void)
 {
   void *ptr;
   diff_buffer = page_aligned_alloc (&ptr, record_size);
+  if (listed_incremental_option)
+    read_directory_file ();
 }
 
 /* Sigh about something that differs by writing a MESSAGE to stdlis,
@@ -112,8 +103,7 @@ process_rawdata (size_t bytes, char *buffer)
 
   if (memcmp (buffer, diff_buffer, bytes))
     {
-      report_difference (&current_stat_info,
-			 _("Contents differ"));
+      report_difference (&current_stat_info, _("Contents differ"));
       return 0;
     }
 
@@ -142,14 +132,15 @@ process_dumpdir (size_t bytes, char *buffer)
    address of the chunk it can work with.  The PROCESSOR should return
    nonzero for success.  It it return error once, continue skipping
    without calling PROCESSOR anymore.  */
+
 static void
-read_and_process (off_t size, int (*processor) (size_t, char *))
+read_and_process (struct tar_stat_info *st, int (*processor) (size_t, char *))
 {
   union block *data_block;
   size_t data_size;
-
-  if (multi_volume_option)
-    save_sizeleft = size;
+  size_t size = st->stat.st_size;
+  
+  mv_begin (st);
   while (size)
     {
       data_block = find_next_block ();
@@ -167,9 +158,9 @@ read_and_process (off_t size, int (*processor) (size_t, char *))
       set_next_block_after ((union block *)
 			    (data_block->buffer + data_size - 1));
       size -= data_size;
-      if (multi_volume_option)
-	save_sizeleft -= data_size;
+      mv_size_left (size);
     }
+  mv_end ();
 }
 
 /* Call either stat or lstat over STAT_DATA, depending on
@@ -193,13 +184,255 @@ get_stat_data (char const *file_name, struct stat *stat_data)
   return 1;
 }
 
+
+static void
+diff_dir (void)
+{
+  struct stat stat_data;
+
+  if (!get_stat_data (current_stat_info.file_name, &stat_data))
+    return;
+
+  if (!S_ISDIR (stat_data.st_mode))
+    report_difference (&current_stat_info, _("File type differs"));
+  else if ((current_stat_info.stat.st_mode & MODE_ALL) !=
+	   (stat_data.st_mode & MODE_ALL))
+    report_difference (&current_stat_info, _("Mode differs"));
+}
+
+static void
+diff_file (void)
+{
+  char const *file_name = current_stat_info.file_name;
+  struct stat stat_data;
+
+  if (!get_stat_data (file_name, &stat_data))
+    skip_member ();
+  else if (!S_ISREG (stat_data.st_mode))
+    {
+      report_difference (&current_stat_info, _("File type differs"));
+      skip_member ();
+    }
+  else
+    {
+      if ((current_stat_info.stat.st_mode & MODE_ALL) !=
+	  (stat_data.st_mode & MODE_ALL))
+	report_difference (&current_stat_info, _("Mode differs"));
+
+      if (!sys_compare_uid (&stat_data, &current_stat_info.stat))
+	report_difference (&current_stat_info, _("Uid differs"));
+      if (!sys_compare_gid (&stat_data, &current_stat_info.stat))
+	report_difference (&current_stat_info, _("Gid differs"));
+
+      if (tar_timespec_cmp (get_stat_mtime (&stat_data), 
+                            current_stat_info.mtime))
+	report_difference (&current_stat_info, _("Mod time differs"));
+      if (current_header->header.typeflag != GNUTYPE_SPARSE
+	  && stat_data.st_size != current_stat_info.stat.st_size)
+	{
+	  report_difference (&current_stat_info, _("Size differs"));
+	  skip_member ();
+	}
+      else
+	{
+	  int atime_flag =
+	    (atime_preserve_option == system_atime_preserve
+	     ? O_NOATIME
+	     : 0);
+
+	  diff_handle = open (file_name, O_RDONLY | O_BINARY | atime_flag);
+
+	  if (diff_handle < 0)
+	    {
+	      open_error (file_name);
+	      skip_member ();
+	      report_difference (&current_stat_info, NULL);
+	    }
+	  else
+	    {
+	      int status;
+
+	      if (current_stat_info.is_sparse)
+		sparse_diff_file (diff_handle, &current_stat_info);
+	      else
+		read_and_process (&current_stat_info, process_rawdata);
+
+	      if (atime_preserve_option == replace_atime_preserve)
+		{
+		  struct timespec ts[2];
+		  ts[0] = get_stat_atime (&stat_data);
+		  ts[1] = get_stat_mtime (&stat_data);
+		  if (set_file_atime (diff_handle, file_name, ts) != 0)
+		    utime_error (file_name);
+		}
+
+	      status = close (diff_handle);
+	      if (status != 0)
+		close_error (file_name);
+	    }
+	}
+    }
+}
+
+static void
+diff_link (void)
+{
+  struct stat file_data;
+  struct stat link_data;
+
+  if (get_stat_data (current_stat_info.file_name, &file_data)
+      && get_stat_data (current_stat_info.link_name, &link_data)
+      && !sys_compare_links (&file_data, &link_data))
+    report_difference (&current_stat_info,
+		       _("Not linked to %s"),
+		       quote (current_stat_info.link_name));
+}
+
+#ifdef HAVE_READLINK
+static void
+diff_symlink (void)
+{
+  size_t len = strlen (current_stat_info.link_name);
+  char *linkbuf = alloca (len + 1);
+
+  int status = readlink (current_stat_info.file_name, linkbuf, len + 1);
+
+  if (status < 0)
+    {
+      if (errno == ENOENT)
+	readlink_warn (current_stat_info.file_name);
+      else
+	readlink_error (current_stat_info.file_name);
+      report_difference (&current_stat_info, NULL);
+    }
+  else if (status != len
+	   || strncmp (current_stat_info.link_name, linkbuf, len) != 0)
+    report_difference (&current_stat_info, _("Symlink differs"));
+}
+#endif
+
+static void
+diff_special (void)
+{
+  struct stat stat_data;
+
+  /* FIXME: deal with umask.  */
+
+  if (!get_stat_data (current_stat_info.file_name, &stat_data))
+    return;
+
+  if (current_header->header.typeflag == CHRTYPE
+      ? !S_ISCHR (stat_data.st_mode)
+      : current_header->header.typeflag == BLKTYPE
+      ? !S_ISBLK (stat_data.st_mode)
+      : /* current_header->header.typeflag == FIFOTYPE */
+      !S_ISFIFO (stat_data.st_mode))
+    {
+      report_difference (&current_stat_info, _("File type differs"));
+      return;
+    }
+
+  if ((current_header->header.typeflag == CHRTYPE
+       || current_header->header.typeflag == BLKTYPE)
+      && current_stat_info.stat.st_rdev != stat_data.st_rdev)
+    {
+      report_difference (&current_stat_info, _("Device number differs"));
+      return;
+    }
+
+  if ((current_stat_info.stat.st_mode & MODE_ALL) !=
+      (stat_data.st_mode & MODE_ALL))
+    report_difference (&current_stat_info, _("Mode differs"));
+}
+
+static void
+diff_dumpdir (void)
+{
+  char *dumpdir_buffer;
+  dev_t dev = 0;
+  struct stat stat;
+
+  if (deref_stat (true, current_stat_info.file_name, &stat))
+    {
+      if (errno == ENOENT)
+	stat_warn (current_stat_info.file_name);
+      else
+	stat_error (current_stat_info.file_name);
+    }
+  else
+    dev = stat.st_dev;
+
+  dumpdir_buffer = get_directory_contents (current_stat_info.file_name, dev);
+
+  if (dumpdir_buffer)
+    {
+      dumpdir_cursor = dumpdir_buffer;
+      read_and_process (&current_stat_info, process_dumpdir);
+      free (dumpdir_buffer);
+    }
+  else
+    read_and_process (&current_stat_info, process_noop);
+}
+
+static void
+diff_multivol (void)
+{
+  struct stat stat_data;
+  int fd, status;
+  off_t offset;
+
+  if (current_stat_info.had_trailing_slash)
+    {
+      diff_dir ();
+      return;
+    }
+
+  if (!get_stat_data (current_stat_info.file_name, &stat_data))
+    return;
+
+  if (!S_ISREG (stat_data.st_mode))
+    {
+      report_difference (&current_stat_info, _("File type differs"));
+      skip_member ();
+      return;
+    }
+
+  offset = OFF_FROM_HEADER (current_header->oldgnu_header.offset);
+  if (stat_data.st_size != current_stat_info.stat.st_size + offset)
+    {
+      report_difference (&current_stat_info, _("Size differs"));
+      skip_member ();
+      return;
+    }
+
+  fd = open (current_stat_info.file_name, O_RDONLY | O_BINARY);
+
+  if (fd < 0)
+    {
+      open_error (current_stat_info.file_name);
+      report_difference (&current_stat_info, NULL);
+      skip_member ();
+      return;
+    }
+
+  if (lseek (fd, offset, SEEK_SET) < 0)
+    {
+      seek_error_details (current_stat_info.file_name, offset);
+      report_difference (&current_stat_info, NULL);
+      return;
+    }
+
+  read_and_process (&current_stat_info, process_rawdata);
+  
+  status = close (fd);
+  if (status != 0)
+    close_error (current_stat_info.file_name);
+}
+
 /* Diff a file against the archive.  */
 void
 diff_archive (void)
 {
-  struct stat stat_data;
-  int status;
-  struct utimbuf restore_times;
 
   set_next_block_after (current_header);
   decode_header (current_header, &current_stat_info, &current_format, 1);
@@ -216,7 +449,7 @@ diff_archive (void)
   switch (current_header->header.typeflag)
     {
     default:
-      ERROR ((0, 0, _("%s: Unknown file type '%c', diffed as normal file"),
+      ERROR ((0, 0, _("%s: Unknown file type `%c', diffed as normal file"),
 	      quotearg_colon (current_stat_info.file_name),
 	      current_header->header.typeflag));
       /* Fall through.  */
@@ -229,265 +462,40 @@ diff_archive (void)
       /* Appears to be a file.  See if it's really a directory.  */
 
       if (current_stat_info.had_trailing_slash)
-	goto really_dir;
-
-      if (!get_stat_data (current_stat_info.file_name, &stat_data))
-	{
-	  skip_member ();
-	  goto quit;
-	}
-
-      if (!S_ISREG (stat_data.st_mode))
-	{
-	  report_difference (&current_stat_info, _("File type differs"));
-	  skip_member ();
-	  goto quit;
-	}
-
-      if ((current_stat_info.stat.st_mode & MODE_ALL) !=
-	  (stat_data.st_mode & MODE_ALL))
-	report_difference (&current_stat_info, _("Mode differs"));
-
-      if (!sys_compare_uid (&stat_data, &current_stat_info.stat))
-	report_difference (&current_stat_info, _("Uid differs"));
-      if (!sys_compare_gid (&stat_data, &current_stat_info.stat))
-	report_difference (&current_stat_info, _("Gid differs"));
-
-      if (stat_data.st_mtime != current_stat_info.stat.st_mtime)
-	report_difference (&current_stat_info, _("Mod time differs"));
-      if (current_header->header.typeflag != GNUTYPE_SPARSE &&
-	  stat_data.st_size != current_stat_info.stat.st_size)
-	{
-	  report_difference (&current_stat_info, _("Size differs"));
-	  skip_member ();
-	  goto quit;
-	}
-
-      diff_handle = open (current_stat_info.file_name, O_RDONLY | O_BINARY);
-
-      if (diff_handle < 0)
-	{
-	  open_error (current_stat_info.file_name);
-	  skip_member ();
-	  report_difference (&current_stat_info, NULL);
-	  goto quit;
-	}
-
-      restore_times.actime = stat_data.st_atime;
-      restore_times.modtime = stat_data.st_mtime;
-
-      /* Need to treat sparse files completely differently here.  */
-
-      if (current_stat_info.is_sparse)
-	sparse_diff_file (diff_handle, &current_stat_info);
+	diff_dir ();
       else
-	{
-	  if (multi_volume_option)
-	    {
-	      assign_string (&save_name, current_stat_info.file_name);
-	      save_totsize = current_stat_info.stat.st_size;
-	      /* save_sizeleft is set in read_and_process.  */
-	    }
-
-	  read_and_process (current_stat_info.stat.st_size, process_rawdata);
-
-	  if (multi_volume_option)
-	    assign_string (&save_name, 0);
-	}
-
-      status = close (diff_handle);
-      if (status != 0)
-	close_error (current_stat_info.file_name);
-
-      if (atime_preserve_option)
-	utime (current_stat_info.file_name, &restore_times);
-
-    quit:
+	diff_file ();
       break;
 
     case LNKTYPE:
-      {
-	struct stat file_data;
-	struct stat link_data;
-
-	if (!get_stat_data (current_stat_info.file_name, &file_data))
-	  break;
-	if (!get_stat_data (current_stat_info.link_name, &link_data))
-	  break;
-	if (!sys_compare_links (&file_data, &link_data))
-	  report_difference (&current_stat_info,
-			     _("Not linked to %s"),
-			     quote (current_stat_info.link_name));
-      }
+      diff_link ();
       break;
 
 #ifdef HAVE_READLINK
     case SYMTYPE:
-      {
-	size_t len = strlen (current_stat_info.link_name);
-	char *linkbuf = alloca (len + 1);
-
-	status = readlink (current_stat_info.file_name, linkbuf, len + 1);
-
-	if (status < 0)
-	  {
-	    if (errno == ENOENT)
-	      readlink_warn (current_stat_info.file_name);
-	    else
-	      readlink_error (current_stat_info.file_name);
-	    report_difference (&current_stat_info, NULL);
-	  }
-	else if (status != len
-		 || strncmp (current_stat_info.link_name, linkbuf, len) != 0)
-	  report_difference (&current_stat_info, _("Symlink differs"));
-
-	break;
-      }
+      diff_symlink ();
+      break;
 #endif
 
     case CHRTYPE:
     case BLKTYPE:
     case FIFOTYPE:
-
-      /* FIXME: deal with umask.  */
-
-      if (!get_stat_data (current_stat_info.file_name, &stat_data))
-	break;
-
-      if (current_header->header.typeflag == CHRTYPE
-	  ? !S_ISCHR (stat_data.st_mode)
-	  : current_header->header.typeflag == BLKTYPE
-	  ? !S_ISBLK (stat_data.st_mode)
-	  : /* current_header->header.typeflag == FIFOTYPE */
-	  !S_ISFIFO (stat_data.st_mode))
-	{
-	  report_difference (&current_stat_info, _("File type differs"));
-	  break;
-	}
-
-      if ((current_header->header.typeflag == CHRTYPE
-	   || current_header->header.typeflag == BLKTYPE)
-	  && current_stat_info.stat.st_rdev != stat_data.st_rdev)
-	{
-	  report_difference (&current_stat_info, _("Device number differs"));
-	  break;
-	}
-
-      if ((current_stat_info.stat.st_mode & MODE_ALL) != (stat_data.st_mode & MODE_ALL))
-	{
-	  report_difference (&current_stat_info, _("Mode differs"));
-	  break;
-	}
-
+      diff_special ();
       break;
 
     case GNUTYPE_DUMPDIR:
-      {
-	char *dumpdir_buffer = get_directory_contents (current_stat_info.file_name, 0);
-
-	if (multi_volume_option)
-	  {
-	    assign_string (&save_name, current_stat_info.file_name);
-	    save_totsize = current_stat_info.stat.st_size;
-	    /* save_sizeleft is set in read_and_process.  */
-	  }
-
-	if (dumpdir_buffer)
-	  {
-	    dumpdir_cursor = dumpdir_buffer;
-	    read_and_process (current_stat_info.stat.st_size, process_dumpdir);
-	    free (dumpdir_buffer);
-	  }
-	else
-	  read_and_process (current_stat_info.stat.st_size, process_noop);
-
-	if (multi_volume_option)
-	  assign_string (&save_name, 0);
-	/* Fall through.  */
-      }
+      diff_dumpdir ();
+      /* Fall through.  */
 
     case DIRTYPE:
-    really_dir:
-      if (!get_stat_data (current_stat_info.file_name, &stat_data))
-	break;
-
-      if (!S_ISDIR (stat_data.st_mode))
-	{
-	  report_difference (&current_stat_info, _("File type differs"));
-	  break;
-	}
-
-      if ((current_stat_info.stat.st_mode & MODE_ALL) != (stat_data.st_mode & MODE_ALL))
-	{
-	  report_difference (&current_stat_info, _("Mode differs"));
-	  break;
-	}
-
+      diff_dir ();
       break;
 
     case GNUTYPE_VOLHDR:
       break;
 
     case GNUTYPE_MULTIVOL:
-      {
-	off_t offset;
-
-	if (current_stat_info.had_trailing_slash)
-	  goto really_dir;
-
-	if (!get_stat_data (current_stat_info.file_name, &stat_data))
-	  break;
-
-	if (!S_ISREG (stat_data.st_mode))
-	  {
-	    report_difference (&current_stat_info, _("File type differs"));
-	    skip_member ();
-	    break;
-	  }
-
-	offset = OFF_FROM_HEADER (current_header->oldgnu_header.offset);
-	if (stat_data.st_size != current_stat_info.stat.st_size + offset)
-	  {
-	    report_difference (&current_stat_info, _("Size differs"));
-	    skip_member ();
-	    break;
-	  }
-
-	diff_handle = open (current_stat_info.file_name, O_RDONLY | O_BINARY);
-
-	if (diff_handle < 0)
-	  {
-	    open_error (current_stat_info.file_name);
-	    report_difference (&current_stat_info, NULL);
-	    skip_member ();
-	    break;
-	  }
-
-	if (lseek (diff_handle, offset, SEEK_SET) < 0)
-	  {
-	    seek_error_details (current_stat_info.file_name, offset);
-	    report_difference (&current_stat_info, NULL);
-	    break;
-	  }
-
-	if (multi_volume_option)
-	  {
-	    assign_string (&save_name, current_stat_info.file_name);
-	    save_totsize = stat_data.st_size;
-	    /* save_sizeleft is set in read_and_process.  */
-	  }
-
-	read_and_process (current_stat_info.stat.st_size, process_rawdata);
-
-	if (multi_volume_option)
-	  assign_string (&save_name, 0);
-
-	status = close (diff_handle);
-	if (status != 0)
-	  close_error (current_stat_info.file_name);
-
-	break;
-      }
+      diff_multivol ();
     }
 }
 
